import argparse
import base64
import json
import logging
import os
import time
import typing

from typing import List

from .base import Ec2BaseAction

class Launch(Ec2BaseAction):
    """ Action class for launching ec2 spot instances. """

    def _parse_args(self, parser, args):
        parser.add_argument('-n', '--name', type=str, required=True,
                            help='name for instance')
        parser.add_argument('-s', '--specification', type=str, required=True,
                            help='path to specification JSON file')
        parser.add_argument('-p', '--price', type=str, required=True,
                            help='max bid price for instance')
        parser.add_argument('-u', '--userdata', type=str, default=None,
                            help='optional path to userdata shell script')

        return parser.parse_args(args)

    def execute(self):
        """ Launch ec2 spot instance, store information by name. """
        # load launch specifications and base64 encoded userdata
        with open(self.args.specification) as f:
            launch_spec = json.loads(f.read())
        if self.args.userdata is not None:
            with open(self.args.userdata) as f:
                userdata_str = f.read()
                userdata_bytes = base64.b64encode(bytes(userdata_str, 'utf-8'))
                userdata = userdata_bytes.decode('utf-8')
                launch_spec['UserData'] = userdata

        # request instance
        request = self.client.request_spot_instances(
            LaunchSpecification=launch_spec,
            SpotPrice=self.args.price
        )
        request_id = request['SpotInstanceRequests'][0].get('SpotInstanceRequestId')

        if request_id is None:
            print('Request submission failed.')
            return False
        else:
            print(f"\nSpot instance request submitted.\nSpotInstanceRequestId: {request_id}\n")

        # check request state every 1.5 seconds, get instance id when fulfilled
        holding = ['pending-evaluation','not-scheduled-yet',
                   'pending-fulfillment']
        request_status = 'pending-evaluation'
        while request_status in holding:
            request_state = self.client.describe_spot_instance_requests(
                SpotInstanceRequestIds=[request_id]
            )['SpotInstanceRequests'][0]
            request_status = request_state['Status'].get('Code')
            time.sleep(1.5)

        # handle successful fulfillment
        if request_status == 'fulfilled':
            instance_id = request_state.get('InstanceId')
            print(f"Spot instance request fulfilled.\nInstanceId: {instance_id}\n")
            # Add name tag to resource
            self.client.create_tags(
                Resources=[instance_id],
                Tags=[{'Key': 'Name',
                       'Value': self.args.name}]
            )

            return True

        # handle failed fulfillment
        else:
            self.client.cancel_spot_instance_requests(
                SpotInstanceRequestIds=[request_id]
            )
            print(f"Spot instance request cancelled.\nReason: {request_status}\n")

            return False


from typing import List

from ..base import BaseAction

class Ec2BaseAction(BaseAction):
    """ Base class for ec2 actions. """

    def _find_instance_by_name(self, name: str):
        """ Lookup resource by name """
        # TODO: Handle case where multiple instances have same name
        instances = self._get_active_instances()
        for instance in instances:
            if self._get_instance_name(instance) == name:
                return instance

    def _get_instance_name(self, instance: List)->str:
        """ Get instance name from instance description """
        tags = instance['Tags']
        for tag in tags:
            if tag['Key'] == 'Name':
                return tag['Value']

    def _get_active_instances(self)->List:
        """ Get list of running instances """
        filters = {'instance-lifecycle': ['spot'],
                   'instance-state-name': ['running']}
        filters_list = [{'Name': k, 'Values': v} for k, v in filters.items()]

        response = self.client.describe_instances(Filters=filters_list)
        reservations = response.get('Reservations')
        instances = [i for r in reservations for i in r.get('Instances')]

        return instances

    def execute(self, **kwargs):
        """ Method to run action """
        raise NotImplementedError



class BaseAction(object):
    """ Base class for resource actions. """

    def __init__(self, client=None, parser=None, args=None):
        if (client is None) or (parser is None) or (args is None):
            raise ValueError('Action instance missing required arg.')

        self.client = client
        self.args = self._parse_args(parser, args)

    def _parse_args(self, parser, args):
        """ Method to add and parse action specific args """
        raise NotImplementedError

    def run(self, **kwargs):
        """ Method to run action """
        raise NotImplementedError


import base64
import json
import logging
import os
import subprocess
import time
import typing

from typing import List

from .base import Ec2BaseAction

class Ssh(Ec2BaseAction):
    """ Class for managing ec2 spot instances. """

    def _parse_args(self, parser, args):
        parser.add_argument('-n', '--name', type=str, required=True,
                            help='name for instance')
        parser.add_argument('-i', '--identity_file', type=str, required=True,
                            help='path to .pem file')
        parser.add_argument('-u', '--user', type=str, required=True,
                            help='user to login as')
        parser.add_argument('--forward_agent', action='store_true',
                            help='forward SSH agent from local machine')
        parser.add_argument('--create_ssh_profile', action='store_true',
                            help='create profile for awspot ssh config file')
        parser.add_argument('--remove_ssh_profile', action='store_true',
                            help='remove profile for awspot ssh config file')

        return parser.parse_args()

    def create_ssh_profile(self, public_dns: str):
        """ Add profile to ssh config file. """
        config_path = os.path.expanduser("~") + "/.awspot/ssh_config"
        profile = f"\nHost {self.args.name}\n  HostName {public_dns}"
        profile += f"\n  User {self.args.user}\n  IdentityFile {self.args.identity_file}\n"
        if self.args.forward_agent:
            profile += f"\n ForwardAgent Yes\n"

        with open(config_path, 'a') as f:
            f.write(profile)

        print(f'\nProfile for {self.args.name} successfully created.\n')


    def remove_ssh_profile(self):
        config_path = os.path.expanduser("~") + "/.awspot/ssh_config"

        # get lines
        with open(config_path, 'r') as f:
            lines = f.readlines()

        # iterate over lines
        with open(config_path, 'w') as f:
            ignore = False
            for line in lines:
                # ignore lines from target Host til next host or EOF
                if f"Host {self.args.name}" in line:
                    ignore = True
                    continue

                if ignore:
                    if line[:5]=="Host ":
                        ignore = False
                    else:
                        continue
            
                f.write(line)

    def execute(self):
        # TODO: Automatically detect if profile exists for instance by name
        #       Use if so. Removes requirement for -i and -u args 
        instance = self._find_instance_by_name(self.args.name)
        public_dns = instance['PublicDnsName']

        if self.args.create_ssh_profile:
            self.create_ssh_profile(public_dns)
        elif self.args.remove_ssh_profile:
            self.remove_ssh_profile()
        else:
            subprocess.run(["ssh", "-i", self.args.identity_file,
                            f"{self.args.user}@{public_dns}"])


from .base import Ec2BaseAction

class ListActive(Ec2BaseAction):
    """ Action for listing active ec2 spot instances. """

    def _parse_args(self, parser, args):
        return None

    def execute(self):
        """ List active instances of resource. """
        instances = self._get_active_instances()
        if not instances:
            print("No running instances.")
        else:
            output = "\nInstanceId\t\tName\n\n"
            for instance in instances:
                name = self._get_instance_name(instance)
                instance_id = instance['InstanceId']
                output += f"{instance_id}\t{name}\n"

            print(output)


import json
import typing

class Manager(object):
    """ Base class for spot resource managers."""

    def __init__(self, session, parser, args):
        self.client = self._initialize_client(session)
        self.parser = parser
        self.args = args

    def _initialize_client(self, session):
        """ Initialize client for resource """
        raise NotImplementedError

    def launch(self, **kwargs):
        """ Launch resource. """
        raise NotImplementedError

    def list_running(self, **kwargs):
        """ List running instances of resource. """
        raise NotImplementedError

    def terminate(self, **kwargs):
        """ Terminate resource by name. """
        raise NotImplementedError

    def execute(self):
        """ Execute appropriate action """
        raise NotImplementedError


import typing

from .base import Manager

class ec2Manager(Manager):
    """ Class for managing ec2 spot instances. """

    def _initialize_client(self, session):
        return session.client('ec2')

    def launch(self):
        from awspot.actions.ec2 import Launch
        action = Launch(self.client, self.parser, self.args)
        return action.execute()

    def terminate(self):
        from awspot.actions.ec2 import Terminate
        action = Terminate(self.client, self.parser, self.args)
        return action.execute()

    def list_active(self):
        from awspot.actions.ec2 import ListActive
        action = ListActive(self.client, self.parser, self.args)
        return action.execute()

    def ssh(self):
        from awspot.actions.ec2 import Ssh
        action = Ssh(self.client, self.parser, self.args)
        return action.execute()
    
    def execute(self, action: str):
        method = getattr(self, action)
        return method()


#!/usr/bin/env python
from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='AWSpot',
    version='0.0.5',
    author='Rob Dalton',
    author_email='rob@robdalton.me',
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3.6"
    ],
    packages=find_packages(),
    scripts=['bin/awspot'],
    url='http://pypi.python.org/pypi/AWSpot/',
    license='LICENSE',
    description='Utility for managing AWS spot resources.',
    long_description=open('README.md').read(),
    install_requires=[
        "boto3 >= 1.7.24"
    ],
)



import argparse

def get_base_argparser():
    # setup default argument parser (will be passed down to scripts)
    parser = argparse.ArgumentParser(description='Script to manage AWS spot resources') 
    parser.add_argument('resource_type', type=str,
                        choices=['ec2', 'emr'],
                        help='resource type to manage')
    parser.add_argument('action', type=str,
                        choices=['launch', 'list_active', 'terminate', 'ssh'],
                        help='action type')
    parser.add_argument('--profile', type=str,
                        help='AWS profile to use', default='default')

    return parser



import boto3
import sys

from unittest import TestCase
from unittest.mock import Mock, patch

from awspot.etc import get_base_argparser
from awspot.managers import ec2Manager

class TestEc2Manager(TestCase):
    # TODO: Find isolated way to mock boto3 calls. Moto does not provide enough
    #       functionality.

    MOCK_CLIENT_ATTRS = {
        'request_spot_instances.return_value': {
            'SpotInstanceRequests': [{'SpotInstanceRequestId': 'sir-fa789d54'}]
        },
        'describe_spot_instance_requests.return_value': {
            'SpotInstanceRequests': [{'InstanceId': 'i-0578da97c8ee56262', 'Status': {'Code': 'fulfilled'}}]
        },
        'describe_instances.return_value': {
            'Reservations': [{
                    'Instances': [{
                            'InstanceId': 'i-0578da97c8ee56262',
                            'Tags': [{'Key': 'Name', 'Value': 'test'}]
                        }]
                }]
        },
        'terminate_instances.return_value': {
            'TerminatingInstances': [{'InstanceId': 'i-0578da97c8ee56262'}]
        }
    }

    @classmethod
    def setUpClass(cls):
        mock_client = Mock()
        mock_client.configure_mock(**cls.MOCK_CLIENT_ATTRS)

        mock_session = Mock()
        mock_session.configure_mock(**{'client.return_value': mock_client})

        cls.session = mock_session
        
    def test_launch(self):
        args = [
            'ec2', 'launch',
            '-n', 'test',
            '-s', './assets/ec2_launch_specs.json',
            '-u', './assets/ec2_userdata.sh',
            '-p', '0.070'
        ]
        manager = ec2Manager(self.session, get_base_argparser(), args)
        ret = manager.execute('launch')
        self.assertTrue(ret) 

    def test_terminate(self):
        args = [
            'ec2', 'terminate',
            '-n', 'test'
        ]
        manager = ec2Manager(self.session, get_base_argparser(), args)
        ret = manager.execute('terminate')
        self.assertTrue(ret)


import argparse
import json

from .base import Ec2BaseAction

class Terminate(Ec2BaseAction):
    """ Action for terminating ec2 spot instances. """

    def _parse_args(self, parser, args):
        # TODO: Add handling for case where multiple instances w/same name
        parser.add_argument('-n', '--name', type=str, required=True,
                            help='name for instance')
        return parser.parse_args(args)

    def execute(self):
        """ Terminate resource by name. """
        # TODO: Create fail-safe termination check.
        instance = self._find_instance_by_name(self.args.name)
        if instance:
            instance_id = instance['InstanceId']
            response = self.client.terminate_instances(InstanceIds=[instance_id])
            if response['TerminatingInstances'] and \
               response['TerminatingInstances'][0]['InstanceId'] == instance_id:
                print(f"\nInstance terminating:\n\n  Name: {self.args.name}\n  InstanceId: {instance_id}\n")
                return True

        return False